
#ifndef _MyVector_H_
#define _MyVector_H_

// uvedomil jsem si, ze jsme zadani rikali jinak, nez jsme to pak naprogramovali.. Ale to uz je ted
// jedno :) Zadani jsem rikal tak, ze kdyz nekdo zaindexuje vektor s velkym indexem, ktery tam neni, tak
// se vektor automaticky protahne, nas vyhazuje vyjimku. Uprava je vcelku jednoducha, nechavam ji na vas..


// pomocna funkce, ktera vynuluje pamet zacinajici na ukazateli p a ktera ma velikost bytu
void NulujPamet(void * p, unsigned int velikost)
{
	// pretypujeme void * na char *, do void * nemuzeme nic davat, ani ho posunovat, protoze neni
	// jasne, jakou ma velikost
	char * pch = (char *) p;

	while (velikost-- > 0)
	{
		*pch = 0;
		pch++;
	}
}


// vyjimka, ktera se bude vyhazovat kdyz bude spatny index
// nemusi mit zadna data, to ze se vyhodi, je dostatecna informace o chybe
class SpatnyIndexException
{
};

// a jeste jedna pro neplatnou velikost vektoru
class NeplatnaVelikostException
{
};

// dopredna deklarace iteratoru, protoze MyVector potrebuje k deklaraci i iterator, tak aby
// to nekricelo, ze MyVectorIterator je neznamy
template <class T>
class MyVectorIterator;

// -----------------------------------------------------------------------------
// vlastni sablona naseho vectoru
//

template <class T>
class MyVector
{
	// aby iterator mohl sahat primo na data a nemusel zprostredkovane pres metody MyVector,
	// dame ho jako friend
	friend class MyVectorIterator;

public:
	// konstruktoru se zada inicialni velikost
	MyVector(int velikost);
	// dynamicky alokujeme, potrebujeme destruktor
	~MyVector();

	// zmensi nebo zvetsi vektor
	void ZmenVelikost(int velikost);
	// aby se dal vektor pouzivat stejne jako pole, pretizime operator[]
	// vraci referenci na T, aby se dalo pomoci [] i zapisovat. Bez & by se vracela kopie
	// a neslo by zapsat rovnou do vektoru
	T & operator[](int index);

	// vrati iterator na prvni prvek
	MyVectorIterator<T> begin();
	// vrati iterator za posledni prvek
	MyVectorIterator<T> end();
	
	// tento typedef zpusobi, ze nikoho nemusi zajimat, jak se presne jmenuje trida slouzici
	// jako iterator, proste muze pouzivat napr. kdyz ma vektor intu
	// MyVector<int>::iterator
	typedef MyVectorIterator<T> iterator;

private:
	// ukazatel na dynamicky naalokovanou pamet s daty (polem prvku typu T)
	T * data;
	// kolik mame naalokovano prave polozek
	int velikost;
};


//---------------------------------------------------------------------------
// implementace metod MyVector
//

template <class T>
MyVector<T>::MyVector(int velikost)
{
	// naalokujeme a snulujeme pamet
	data = new T[velikost];
	NulujPamet(data, velikost * sizeof(T));
	// zapamatujeme si velikost co jsme naalokovali
	this->velikost = velikost;
}

template <class T>
MyVector<T>::~MyVector()
{
	// alokujeme pomoci new s hranatymi zavorkami, musime uvolnovat s delete s []
	delete[] data;
};

template <class T>
void MyVector<T>::ZmenVelikost(int velikost)
{
	// kdyz by byla nova velikost stejna jako nase aktualni, je zbytecne neco delat
	if (this->velikost == velikost)
		return;
	
	if (velikost < 0) // toto jsme zapomneli na hodine osetrit...
		throw NeplatnaVelikostException();

	// naalokujeme nova data a pokud alokoujeme vic, nez budeme prepisovat starymi daty, tak tento
	// zbytek vynulujeme. To predtim neni treba nulovat, protoze se to stejne prepise z puvodniho
	// pole
	T * nova_data = new T[velikost];
	if (this->velikost < velikost)
		NulujPamet(nova_data + this->velikost, (velikost - this->velikost) * sizeof(T));

	// kopirujeme
	for (int i = 0; i < velikost && i < this->velikost; i++)
	{
		nova_data[i] = data[i];
	}
	// odalokujeme stary buffer
	delete[] data;
	// zapamatujeme si ukazatel na novy buffer a velikost noveho bufferu
	data = nova_data;
	this->velikost = velikost;
}

template <class T>
T & MyVector<T>::operator[](int index)
{
	if (index < 0 || index >= velikost) // kontrola indexu
		throw SpatnyIndexException();		

	// index je spravny, muzeme sahnout do pole
	return data[index];
}

template <class T>
MyVectorIterator<T> MyVector<T>::begin()
{
	// vytvorime novy objekt iteratoru, konstruktoru dame spravne parametry
	return MyVectorIterator<T>(*this, 0);
}

template <class T>
MyVectorIterator<T> MyVector<T>::end()
{
	// velikost je index v poli za konec, coz je spravny prametr pro iterator end()
	return MyVectorIterator<T>(*this, velikost);
}


//---------------------------------------------------------------------------------
// sablona iteratoru pro MyVector
// implementuje dopredny iterator, ktery jde posouvat jen dopredu a jde pomoci nej
// cist i zapisovat
//

template <class T>
class MyVectorIterator
{
public:
	// konstruktor, muze se mu dat jen jeden parametr, pak bude implicitne ukazovat na prvni
	// prvek
	MyVectorIterator(MyVector<T> & pole, int index = 0);

	// prefixova verze ++
	MyVectorIterator operator++();
	// postfixova verze ++
	MyVectorIterator operator++(int);

	// a dva operatory, abychom mohli porovnavat iteratory, podobne jako ukazatele
	bool operator==(const MyVectorIterator & it2);
	bool operator!=(const MyVectorIterator & it2);

	// operator ktery umozni dereferencovat, vraci referenci podobne jako operator[] u MyVector, aby
	// pres nej slo i zapsat
	T & operator*();

private:
	// musime si pamatovat k jakemu poli patrime
	MyVector<T> * pole;
	// a na jakou polozku v tom poli prave ukazujeme
	int index;
};

template <class T>
MyVectorIterator<T>::MyVectorIterator(MyVector<T> & pole, int index)
{
	this->pole = &pole;
	this->index = index;
}

template <class T>
MyVectorIterator<T> MyVectorIterator<T>::operator++() // prefixova verze
{
	if (index == pole->velikost)
		throw SpatnyIndexException(); // jsme za poslednim prvkem a meli bychom se posunout jeste dale..

	// v poradku, posunujeme
	index++;
	return *this;
}

template <class T>
MyVectorIterator<T> MyVectorIterator<T>::operator++(int) // postfixova verze
{
	if (index == pole->velikost)
		throw SpatnyIndexException(); // jsme za poslednim prvkem a meli bychom se posunout jeste dale..

	// budeme vracet iterator pred posunutim, takze si zkopirujeme do pomocne promenne aktualni stav
	MyVectorIterator novy = *this;
	// posunujeme
	index++;
	// a vratime zapamatovany ukazujici o jedno pole zpet
	return novy;
}

template <class T>
T & MyVectorIterator<T>::operator*()
{
	// kdyz bychom nebyli friend s MyVector, mohli bychom volat jen public metodu
	//return (*pole)[index];

	if (index < 0 || index >= pole->velikost)
		throw SpatnyIndexException();		

	// my ale friend jsme, muzeme si sahnout rovnou na data. Ovsem kdyz se zmeni vnitrni struktura
	// MyVector, bude se muset prepsat i tady iterator, to by se pri volani metody nemuselo
	// Toto je ale trosku rychlejsi.
	return pole->data[index];
}

template <class T>
bool MyVectorIterator<T>::operator==(const MyVectorIterator & i2)
{
	// dva iteratory se rovnaji pokud ukazuji do stejneho pole a na stejny prvek
	if (pole == i2.pole && index == i2.index)
		return true;
	else
		return false;
}

template <class T>
bool MyVectorIterator<T>::operator!=(const MyVectorIterator & i2)
{
	// operator != dela presne opak operatoru ==
	return !operator==(i2);
}


#endif